#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Ming-Ke-Ming
    ~~~~~~~~~~~~

    Common Identity Module for decentralized user identity authentication
"""

from mkm.crypto import SymmetricKey, PrivateKey, PublicKey
from mkm.address import NetworkID, Address
from mkm.meta import Meta
from mkm.entity import ID, Entity

name = "MingKeMing"

__author__ = 'Albert Moky'

__all__ = [
    'SymmetricKey',
    'PrivateKey', 'PublicKey',

    'NetworkID', 'Address', 'ID', 'Meta',
    'Entity',
]
